@@ -20,6 +20,9 @@
 #include "common.h"
 #include "context.h"
 
+// This struct must be the first member of the gpu's priv struct. The `pl_gpu`
+// helpers will cast the priv struct to this struct!
+
 #define GPU_PFN(name) __typeof__(pl_##name) *name
 struct pl_gpu_fns {
     // Destructors: These also free the corresponding objects, but they
@@ -78,15 +81,17 @@ void pl_gpu_print_info(const struct pl_gpu *gpu, enum pl_log_level lev);
 // terms of bit depth / component index.
 void pl_gpu_sort_formats(struct pl_gpu *gpu);
 
+// Perform some basic validity checks on all pl_fmts
+void pl_gpu_verify_formats(struct pl_gpu *gpu);
+
 // Pretty-print the format list
 void pl_gpu_print_formats(const struct pl_gpu *gpu, enum pl_log_level lev);
 
 // Look up the right GLSL image format qualifier from a partially filled-in
 // pl_fmt, or NULL if the format does not have a legal matching GLSL name.
 //
-// Warning: If `fmt->emulated` is true, this function makes the hard assumption
-// that 3-channel formats are being emulated as equivalent 4-channel formats!
-const char *pl_fmt_glsl_format(const struct pl_fmt *fmt);
+// `components` may differ from fmt->num_components (for emulated formats)
+const char *pl_fmt_glsl_format(const struct pl_fmt *fmt, int components);
 
 // Compute the total size (in bytes) of a texture transfer operation
 size_t pl_tex_transfer_size(const struct pl_tex_transfer_params *par);
