@@ -134,7 +134,7 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
         repr->alpha = PL_ALPHA_PREMULTIPLIED;
     }
 
-    GLSL("}");
+    GLSL("}\n");
 }
 
 void pl_shader_encode_color(struct pl_shader *sh,
@@ -554,7 +554,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
     if (!obj)
         return false;
 
-    const struct pl_gpu *gpu = sh->gpu;
+    const struct pl_gpu *gpu = SH_GPU(sh);
     obj->gpu = gpu;
 
     if (!obj->buf) {
@@ -1019,13 +1019,20 @@ void pl_shader_dither(struct pl_shader *sh, int new_depth,
     }
 
     enum pl_dither_method method = params->method;
+    bool can_fixed = sh_glsl(sh).version >= 130;
     ident_t lut = NULL;
     int lut_size = 0;
 
+    if (method == PL_DITHER_ORDERED_FIXED && !can_fixed) {
+        PL_WARN(sh, "PL_DITHER_ORDERED_FIXED requires glsl version >= 130.."
+                " falling back.");
+        goto fallback;
+    }
+
     if (dither_method_is_lut(method)) {
         if (!dither_state) {
-            PL_TRACE(sh, "LUT-based dither method specified but no dither state "
-                     "object given, falling back to non-LUT based methods.");
+            PL_WARN(sh, "LUT-based dither method specified but no dither state "
+                    "object given, falling back to non-LUT based methods.");
             goto fallback;
         }
 
@@ -1048,13 +1055,9 @@ void pl_shader_dither(struct pl_shader *sh, int new_depth,
     goto done;
 
 fallback:
-    if (sh->gpu && sh->gpu->glsl.version >= 130) {
-        method = PL_DITHER_ORDERED_FIXED;
-    } else {
-        method = PL_DITHER_WHITE_NOISE;
-    }
-
+    method = can_fixed ? PL_DITHER_ORDERED_FIXED : PL_DITHER_WHITE_NOISE;
     // fall through
+
 done: ;
 
     int size = 0;
@@ -1069,7 +1072,7 @@ done: ;
         GLSL("vec2 pos = fract(gl_FragCoord.xy * 1.0/%d.0);\n", size);
 
         if (params->temporal) {
-            int phase = sh->index % 8;
+            int phase = SH_PARAMS(sh).index % 8;
             float r = phase * (M_PI / 2); // rotate
             float m = phase < 4 ? 1 : -1; // mirror
             float mat[2][2] = {
