@@ -48,6 +48,7 @@
 
 // Shared struct used to hold vulkan context information
 struct vk_ctx {
+    void *ta; // allocations bound to the lifetime of this vk_ctx
     const struct pl_vk_inst *internal_instance;
     struct pl_context *ctx;
     VkInstance inst;
@@ -57,6 +58,9 @@ struct vk_ctx {
     VkExtent3D transfer_alignment; // for pool_transfer
     VkDevice dev;
 
+    // Generic error flag for catching "failed" devices
+    bool failed;
+
     // Enabled extensions
     const char **exts;
     int num_exts;
@@ -84,7 +88,7 @@ struct vk_ctx {
     // Common pool of signals, to avoid having to re-create these objects often
     struct vk_signal **signals;
     int num_signals;
-    bool events_unsupported;
+    bool disable_events;
 
     // Instance-level function pointers
     VK_FUN(vkGetPhysicalDeviceProperties2KHR);
